"""
Description
===========

This middleware will call a function on, and only on, the first request.
Multiple functions can be registered and will be called one-by-one.

Registering functions after first call will have no effect.

This middleware is best added last to the middlewares that will be called
by your application since a middleware that does not dispatch to another
middleware will inhibit this middleware.


Usage
=====

.. code-block:: python

    from flask import Flask
    from flask_gordon.middleware import BeforeFirstCall


    def once_before_first_call():
        print("Hello World")


    app = Flask(__name__)
    app.wsgi_app = BeforeFirstCall(app.wsgi_app, once_before_first_call)
    app.run()


Classes
=======

.. autoclass:: BeforeFirstCall
  :members: before_first_call

"""
import typing as t

if t.TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment

__all__: t.List[str] = []


class BeforeFirstCall:
    """
    Register function ``fn`` to be called on first WSGI application call.

    Multiple functions can be registered with the :meth:`before_first_call`
    function.

    Parameters
    ----------

    app: WSGIApplication

        The WSGI application to dispatch to after the processing.

    fn: t.Callable

        Function to be called on the first request.
    """

    def __init__(self, app: "WSGIApplication", fn: t.Callable = None) -> None:  # type: ignore[type-arg]
        self.app = app
        self.__functions = set()
        self.__before_first_call = False
        if fn:
            self.__functions.add(fn)

    def __call__(self, environ: "WSGIEnvironment", start_response: "StartResponse") -> t.Iterable[bytes]:
        if not self.__before_first_call:
            for function in self.__functions:
                function()
            self.__before_first_call = True
        return self.app(environ, start_response)

    def before_first_call(self, fn: t.Callable) -> t.Callable:  # type: ignore[type-arg]
        """
        Register a function to be called once on first call

        fn: t.Callable

            Function to be called on the first request.
        """

        self.__functions.add(fn)
        return fn



import typing as t
from copy import deepcopy

from boltons.iterutils import remap


def strip_none(_p: str, k: str, v: t.Any) -> bool:
    return k is not None and v is not None


def _deepmerge(src: t.Dict[str, t.Any], dst: t.Dict[str, t.Any], path: t.List[str] = None) -> t.Dict[str, t.Any]:
    """
    Take every key, value from src and merge it recursively into dst.
    Adapted from https://stackoverflow.com/questions/7204805

    Parameters
    ----------
    dst: dict
        Destination dictionary

    src: dict
        Source dictionary

    path:
        Used internally

    Returns
    -------
    dict:
        Merged dictionary
    """
    if path is None:
        path = []

    for key in src:
        if key not in dst:
            dst[key] = src[key]
        elif isinstance(dst[key], dict) and isinstance(src[key], dict):
            _deepmerge(src[key], dst[key], path + [str(key)])
        elif dst[key] == src[key]:
            pass  # same leaf value
        else:
            dst[key] = src[key]
    return dst


def deepmerge(dst: t.Dict[str, t.Any], src: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    """
    Take every key, value from src and merge it recursivly into dst.
    None values are stripped from src before merging.

    Adapted from https://stackoverflow.com/questions/7204805

    Parameters
    ----------
    dst: dict
        destination dictionary

    src: dict
        source dictionary

    Returns
    -------
    dict:
        merged dictionary
    """
    # pylint: disable=unnecessary-lambda-assignment
    src = remap(src, visit=strip_none)

    return _deepmerge(src, deepcopy(dst))



import typing as t

from torxtools import xdgtools
from torxtools.cfgtools import which
from torxtools.pathtools import expandpath
from yaml import safe_load

from .defaults import CFGFILE_SEARCH_PATHS, make_config


def _read(cfgfile: str) -> t.Dict[str, t.Any]:
    """
    Convenience function in order to be mocked.

    Parameters
    ----------
    cfgfile: str

        a single path representing a yaml file.

    Returns
    -------
    dict:

        a dictionary
    """
    with open(cfgfile, encoding="UTF-8") as fd:
        data = safe_load(fd)
    return data or {}


def prepare_configuration(cfgfilename, cfgfilepaths, default_cfg):
    # Sets environment variables for XDG paths
    xdgtools.setenv()

    # Prepare search path for configuration file. Disable it if we're looking
    # for a impossible (None) file name
    cfgfilename, cfgfilepaths = _prepare_search_paths(cfgfilename, cfgfilepaths)

    # Create a default configuration from what was passed
    # and what we set. Other values are filtered
    defaults = make_config(default_cfg)
    return cfgfilename, cfgfilepaths, defaults


def _prepare_search_paths(cfgfilename, cfgfilepaths):
    # Prepare search path for configuration file. Disable it if we're looking
    # for a impossible (None) file name
    if not cfgfilename:
        cfgfilename = "flask.yml"
    if cfgfilepaths is None:
        cfgfilepaths = CFGFILE_SEARCH_PATHS
    cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]
    return cfgfilename, cfgfilepaths


def read_configuration(cfgfile, cfgfilename, cfgfilepaths):
    # Search for the configuration file
    if cfgfile is None and cfgfilepaths:
        # Search for cfgfile
        cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]
        cfgfile = which(cfgfile, expandpath(cfgfilepaths))

    return _read(cfgfile or "/dev/null") or {}



"""
Description
===========

Initialize a Celery worker or Celery beat for the Flask application.

The `--config` parameter will be parsed, and the yaml contents exported to the Celery app for usage in tasks and beats. The value will be made available in current_app.config["SETTINGS"].


Usage
=====

.. code-block:: python

  #!/usr/bin/env python3

  from flask import Flask
  from flask_gordon.ext import CeleryExt

  flask = Flask(__name__)
  celery = CeleryExt(flask,
      {
          "task_default_queue": "celery.package",
          "task_ignore_result": True,
          "imports": "package.tasks",
      },
  )


.. code-block:: python

  from celery import shared_task
  from flask import current_app

  @shared_task()
  def example():
    print(current_app.config)



Classes
=======

.. autoclass:: CeleryExt
   :members: init_app

"""
import sys
import typing as t

from .._functions import deepmerge
from ..functions import prepare_configuration, read_configuration

try:
    from celery import Celery, Task

    HAS_CELERY = True
except ImportError:
    HAS_CELERY = False


class CeleryExt:
    def __init__(self):
        if not HAS_CELERY:
            raise NotImplementedError("CeleryExt requires celery[redis] package")

    def init_app(
        self,
        app: "Flask",
        configuration=None,
        argv: t.List[str] = None,
        *,
        # borrowed from argparse ext
        cfgfilename=None,
        cfgfilepaths=None,
        default_cfg: t.Dict[str, t.Any] = None,
        key: str = "SETTINGS",
    ):
        """
        Parameters
        ----------
        app: FlaskApp

            A Flask application.

        configuration: dict

            If a configuration dictionnary is passed, it will be used.
            Otherwise, a configuration will be searched for in app.config["CELERY"].

        """

        # pylint: disable=abstract-method
        class FlaskTask(Task):
            def __call__(self, *args: object, **kwargs: t.Dict[str, t.Any]) -> object:
                with app.app_context():
                    return self.run(*args, **kwargs)

        cfgfilename, cfgfilepaths, defaults = prepare_configuration(cfgfilename, cfgfilepaths, default_cfg)

        # Parse arguments if they exist
        if argv is None:
            argv = sys.argv[1:]

        cfgfile = None
        if "--config" in argv:
            idx = argv.index("--config")
            cfgfile = argv[idx + 1]

        data = read_configuration(cfgfile, cfgfilename, cfgfilepaths)

        # # Merge in the reverse order of priority
        cfg = defaults
        cfg = deepmerge(cfg, data)

        configuration = configuration or {}
        if "celery" in cfg:
            cfg["celery"] = deepmerge(cfg["celery"], configuration)
        else:
            cfg["celery"] = configuration

        app.config[key] = cfg

        # Create a celery application
        celery_app = Celery(app.name, task_cls=FlaskTask)
        celery_app.config_from_object(cfg["celery"])
        celery_app.set_default()

        app.extensions["celery"] = celery_app

        return celery_app
